#include <machine/endian.h>
#include <EEPROM.h>

// For one wire temperature measurements
//
#include <OneWire.h>
#include <DallasTemperature.h>

#define COMPOSTA
#include "secrets.h"

// CubeCell Development Framework (Heltec)
// https://github.com/HelTecAutomation/CubeCell-Arduino/releases/download/V1.5.0/package_CubeCell_index.json
//
#include <LoRaWan_APP.h>
#include <CayenneLPP.h>

// Serial will be switched off once steady state is hit.
//
bool debug = true;
#define DEBUG \
  if (debug) Serial

/* LoraWAN, CubeCell and The Things Network (TTN)
   https://heltec-automation-docs.readthedocs.io/en/latest/cubecell/quick_start.html
   https://heltec-automation-docs.readthedocs.io/en/latest/cubecell/lorawan/connect_to_gateway.html
   https://github.com/LukePrior/TTN-BME280-Weather-Station-Heltec-CubeCell-HTCC-AB01

  TTN:  "HTCC-AB01 Class A OTAA"
        EU_863_870
        SF9 for RX2
        JoinEUI 00 00 00 00 00 00 00
        DevEUI generate
        AppKey generate

  Board:  "CubeCell-Board (HTCC-AB01)
  REGION_EU868
  CLASS_A (lower power)
  DevEUI: CUSTOM
  OOTA - should therefore configure things like RX2/SF9,etc automatic.
  ADR = ON (because the device is essentially stationary)
  Linkmode = UNCONFIRMED
  NetReservation = ON (to save join data cross powercycles)
  AT = Off (we do not need this)
  RGB = defactive
  PREAMBLE = 8 (lorawan, default)

  when set LoraWan_RGB to Active,the RGB active in loraWan
   RGB red means sending;
   RGB purple means joined done;
   RGB blue means RxWindow1;
   RGB yellow means RxWindow2;
   RGB green means received done;
*/

// This sensor is wired to Vext - so it can be powered off.
//
#define ONE_WIRE_BUS (GPIO5)

#ifdef ONE_WIRE_BUS
OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature sensors(&oneWire);
#endif

/*LoraWan channelsmask, default channels 0-7*/
uint16_t userChannelsMask[6] = { 0x00FF, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 };

/*LoraWan region, select in arduino IDE tools*/
LoRaMacRegion_t loraWanRegion = ACTIVE_REGION;

/*LoraWan Class, Class A and Class C are supported*/
DeviceClass_t loraWanClass = LORAWAN_CLASS;

/*the application data transmission duty cycle.  value in [ms].*/
const uint32_t appTxDutyCycleMax = 15 * 60 * 1000;  // upload every 15 mins.
// we start evert 10 second; and let this double every time until we hit appTxDutyCycleMax
uint32_t appTxDutyCycle = 10 * 1000;

/* OTAA or ABP*/
bool overTheAirActivation = LORAWAN_NETMODE;

/*ADR enable*/
bool loraWanAdr = LORAWAN_ADR;

/* set LORAWAN_Net_Reserve ON, the node could save the network info to flash, when node reset not need to join again */
bool keepNet = LORAWAN_NET_RESERVE;

/* Indicates if the node is sending confirmed or unconfirmed messages */
bool isTxConfirmed = LORAWAN_UPLINKMODE;

/* Application port */
uint8_t appPort = 42;
/*!
  Number of trials to transmit the frame, if the LoRaMAC layer did not
  receive an acknowledgment. The MAC performs a datarate adaptation,
  according to the LoRaWAN Specification V1.0.2, chapter 18.4, according
  to the following table:

  Transmission nb | Data Rate
  ----------------|-----------
  1 (first)       | DR
  2               | DR
  3               | max(DR-1,0)
  4               | max(DR-1,0)
  5               | max(DR-2,0)
  6               | max(DR-2,0)
  7               | max(DR-3,0)
  8               | max(DR-3,0)

  Note, that if NbTrials is set to 1 or 2, the MAC will not decrease
  the datarate, in case the LoRaMAC layer did not receive an acknowledgment
*/
uint8_t confirmedNbTrials = 4;  // SF8/DR
// uint8_t confirmedNbTrials = 16; // SF9/DR3 ???

/* Prepares the payload of the frame */
static void prepareTxFrame(uint8_t port) {
  CayenneLPP lpp(LORAWAN_APP_DATA_MAX_SIZE);
  int addr = 0;

  lpp.reset();

  DEBUG.print("\nMeasurements:\n");
  // list of options at https://github.com/ElectronicCats/CayenneLPP/blob/master/API.md

  digitalWrite(Vext, LOW);
  delay(50);

#ifdef ONE_WIRE_BUS
  {

    sensors.requestTemperatures();
    delay(800);  // well over 750 mSeconds - needed for the conversion - should use a loraSleep here

    float temp1 = sensors.getTempCByIndex(0);
    float temp2 = sensors.getTempCByIndex(1);
    float temp3 = sensors.getTempCByIndex(2);

    if (debug) {
      DEBUG.print("\tTemperature1 [C]:\t");
      DEBUG.println(temp1);
      DEBUG.print("\tTemperature2 [C]:\t");
      DEBUG.println(temp2);
      DEBUG.print("\tTemperature3 [C]:\t");
      DEBUG.println(temp3);
    };
    lpp.addTemperature(++addr, temp1 /* is in Celcius */);
    lpp.addTemperature(++addr, temp2 /* is in Celcius */);
    lpp.addTemperature(++addr, temp3 /* is in Celcius */);
  };
#endif

  double battV = getBatteryVoltage() / 1000.;
  lpp.addAnalogInput(++addr, battV /* is in mVolt */);

  DEBUG.printf("\tBattery voltage [V]:\t");
  DEBUG.println(battV);

  // lpp.addGPS(1, 4.724002f, 52.036585f, 0.f); // ok met TTN
  // lpp.addVoltage(1, getBatteryVoltage() / 1000. /* is in mVolt */); /// niet ok met TTN does not know this CLPP value
  // lpp.addVoltage(1, 0.50);

  digitalWrite(Vext, HIGH);  // keep power to instruments off if we do not need them.

  appDataSize = lpp.getSize();
  uint8_t *buff = lpp.getBuffer();
  memcpy(appData, buff, appDataSize);

  if (debug) {
    DEBUG.printf("Data (%d bytes <= %d max packet): ", appDataSize, LORAWAN_APP_DATA_MAX_SIZE);
    for (int i = 0; i < appDataSize;) {
      DEBUG.printf(" %02x", buff[i]);
      i++;
      if ((0 == i % 8) || (i == appDataSize)) DEBUG.print("\n");
      if ((0 == i % 8) && (i < appDataSize)) DEBUG.print("                                   ");
    };
  };
}

void setup() {
  Serial.begin(115200);
  Serial.printf("Build: %s\n", rindex(__FILE__, '/'));
  Serial.println("Compiled:" __DATE__ " " __TIME__);

  pinMode(Vext, OUTPUT);
  digitalWrite(Vext, HIGH);  // keep power to instruments off if we do not need them.

#ifdef ONE_WIRE_BUS
  // We skip the intial scan to save energy.
  //
  if (0) {
    digitalWrite(Vext, LOW);
    delay(50);
    sensors.begin();
    digitalWrite(Vext, HIGH);
  };
#endif

  uint64_t i = getID();
#if 1
  DEBUG.printf("ChipID:");
  for (int j = 0; j < sizeof(i); j++)
    DEBUG.printf(" %02X", ((uint8_t *)&i)[sizeof(i) - j - 1]);
  DEBUG.println();
  DEBUG.printf("      : ");
  for (int j = 0; j < sizeof(i); j++)
    DEBUG.printf("0x%02X,", ((uint8_t *)&i)[sizeof(i) - j - 1]);
  DEBUG.println();

  DEBUG.printf("      : 0x%08X%08X\n", (uint32_t)(i >> 32), (uint32_t)i);
#endif

  if (i != chipID) {
    DEBUG.printf("Halting - as this ChipID has not been registered uet.");
    while (1) {};
  }

  deviceState = DEVICE_STATE_INIT;
  LoRaWAN.ifskipjoin();

  Serial.print("Started; state ");
  Serial.println(deviceState);
}

void loop() {
  switch (deviceState) {
    case DEVICE_STATE_INIT:
      {
        DEBUG.println("DEVICE_STATE_INIT");
        LoRaWAN.init(loraWanClass, loraWanRegion);
        deviceState = DEVICE_STATE_JOIN;
        break;
      }
    case DEVICE_STATE_JOIN:
      {
        LoRaWAN.join();
        break;
      }
    case DEVICE_STATE_SEND:
      {
        DEBUG.println("DEVICE_STATE_SEND");

        prepareTxFrame(appPort);
        LoRaWAN.send();
        DEBUG.println("Sending data completed");

        deviceState = DEVICE_STATE_CYCLE;
      }
      break;
    case DEVICE_STATE_CYCLE:
      {
        DEBUG.println("DEVICE_STATE_CYCLE");

        txDutyCycleTime = appTxDutyCycle - APP_TX_DUTYCYCLE_RND / 2 + randr(0, APP_TX_DUTYCYCLE_RND);

        appTxDutyCycle = appTxDutyCycle * 2;
        // Once we hit the nornmal rate - stop there and also
        // reduce logging in the hope that it reduces power.
        if (appTxDutyCycle > appTxDutyCycleMax) {
          appTxDutyCycle = appTxDutyCycleMax;
          debug = false;
        }

        DEBUG.print("Schedule next transmisson in ");
        DEBUG.print(txDutyCycleTime / 1000.);
        DEBUG.println(" seconds.\n");
        DEBUG.flush();

        LoRaWAN.cycle(txDutyCycleTime);

        deviceState = DEVICE_STATE_SLEEP;
        break;
      }
    case DEVICE_STATE_SLEEP:
      {
        LoRaWAN.sleep();
        break;
      }
    default:
      {
        DEBUG.println("DEVICE_STATE_UNKNOWN");
        deviceState = DEVICE_STATE_INIT;
        break;
      }
  }
}
